import argparse

from django.core.management.base import BaseCommand
from django.db import transaction


from content_settings.conf import set_initial_values_for_db


class Command(BaseCommand):
    help = "Create new records and remove unused settings for content settings"

    def add_arguments(self, parser):
        parser.add_argument(
            "--apply", action="store_true", default=False, help="Apply "
        )

    def handle(self, *args, **options):
        with transaction.atomic():
            log = set_initial_values_for_db(apply=options.get("apply"))
            if not log:
                print("No Static Content Settings Updated")
                return

            for l in log:
                print(" ".join(l))

            print("Applied!" if options.get("apply") else "Ignored")



from typing import Any, Optional
from pprint import pformat
from django.core.exceptions import ValidationError

from .validators import call_validator
from . import PREVIEW_NONE


def mix(*cls):
    return type("", cls, {})


class MinMaxValidationMixin:
    min_value = None
    max_value = None

    def validate(self, value):
        super().validate(value)

        is_none_valid = self.min_value is None and self.max_value is None
        if not is_none_valid and value is None:
            raise ValidationError("Value cannot be None")

        if self.min_value is not None and value < self.min_value:
            raise ValidationError(f"Value cannot be less than {self.min_value}")

        if self.max_value is not None and value > self.max_value:
            raise ValidationError(f"Value cannot be greater than {self.max_value}")

    def get_help_format(self):
        yield from super().get_help_format()

        if self.min_value is not None:
            yield f" from {self.min_value}"

        if self.max_value is not None:
            yield f" to {self.max_value}"


class PositiveValidationMixin(MinMaxValidationMixin):
    min_value = 0


class CallToPythonMixin:
    call_func = None
    preview_validators = None
    admin_preview_call = True

    def prepare_python_call(self, value):
        return {"prepared": value}

    def get_preview_validators(self):
        if self.preview_validators is None:
            return ()
        return self.preview_validators

    def get_call_func(self):
        return self.call_func

    def python_call(self, *args, **kwargs):
        return self.call_func(*args, **kwargs)

    def to_python(self, value):
        value = super().to_python(value)
        prepared_kwargs = self.prepare_python_call(value)

        def _(*args, **kwargs):
            return self.python_call(*args, **prepared_kwargs, **kwargs)

        return _

    def give_python_to_admin(self, value, name):
        try:
            value = self.to_python(value)
        except Exception as e:
            return [(None, e)]
        ret = []
        for validator in self.get_preview_validators():
            str_validator = (
                f"{name}({str(validator)})" if self.admin_preview_call else None
            )
            try:
                ret.append((str_validator, validator(value)))
            except Exception as e:
                ret.append((str_validator, e))
        return ret

    def get_admin_preview_html(self, value, name, **kwargs):
        ret = self.give_python_to_admin(value, name)
        if not ret:
            return "No preview (add at least one validator to preview_validators)"
        html = ""
        for validator, val in ret:
            if validator:
                html += f"<div>{validator}</div>"

            if isinstance(val, Exception):
                val = f"ERROR!!! {val}"

            if not self.admin_preview_call and len(ret) == 1:
                html += str(val)
            else:
                html += f"<div>{val}</div>"

        return html

    def get_admin_preview_text(self, value, name, **kwargs):
        ret = self.give_python_to_admin(value, name)
        if not ret:
            return "No preview (add at least one validator to preview_validators)"

        def _():
            for validator, val in ret:
                if validator:
                    yield f">>> {validator}"

                if isinstance(val, Exception):
                    yield f"ERROR!!! {val}"
                elif not self.admin_preview_call and len(ret) == 1:
                    yield str(val)
                else:
                    yield f"<<< {pformat(val)}"

        return "\n".join(_())

    def get_admin_preview_python(self, value, name, **kwargs):
        ret = self.give_python_to_admin(value, name)

        if self.admin_preview_call:
            return ret

        ret = [v for _, v in ret]
        if len(ret) == 1:
            return ret[0]

        return ret


class GiveCallMixin:
    """
    for Template-like classes, when to_python should return a function,
    but as value you want to return not a function, but call it and return result
    """

    admin_preview_call = False

    def get_suffixes(self):
        return ("call",) + super().get_suffixes()

    def get_validators(self):
        return (call_validator(),) + super().get_validators()

    def get_preview_validators(self):
        return (call_validator(),) + super().get_preview_validators()

    def give(self, value, suffix=None):
        if suffix is None:
            return value()
        if suffix == "call":
            return value
        return super().give(value, suffix)


class MakeCallMixin:
    def get_suffixes(self):
        return ("call",) + super().get_suffixes()

    def give_python_to_admin(self, value, name):
        return self.give_python(value)()

    def give(self, value, suffix=None):
        return lambda *args, **kwargs: value


class DictSuffixesMixin:
    suffixes = {}

    def give(self, value, suffix=None):
        if suffix is None:
            return value
        return self.suffixes[suffix](value)


class AdminPreviewMixin:
    def get_admin_preview_html_menu(self, value: str, name: str, **kwargs) -> str:
        return ""

    def get_admin_preview_value(self, value: str, name: str, **kwargs) -> str:
        if self.get_admin_preview_as() == PREVIEW_NONE:
            return ""
        return self.get_admin_preview_html_menu(
            value, name, **kwargs
        ) + super().get_admin_preview_value(value, name, **kwargs)


class AdminPreviewSuffixesMixin(AdminPreviewMixin):
    admin_preview_suffixes = ()
    admin_preview_suffixes_default = "default"

    def get_admin_preview_suffixes_default(self):
        return self.admin_preview_suffixes_default

    def get_admin_preview_suffixes(self, value: str, name: str, **kwargs):
        return self.admin_preview_suffixes

    def get_admin_preview_html_menu(
        self, value: str, name: str, suffix: Optional[str] = None, **kwargs
    ) -> str:
        suffixes = self.get_admin_preview_suffixes(value, name, **kwargs)
        if not len(suffixes):
            return ""

        ret = "<div>"
        for suf in (None, *suffixes):
            if suf == suffix:
                ret += f" <b>{suf or self.get_admin_preview_suffixes_default()}</b> "
            elif suf is None:
                ret += f' <a class="cs_set_params">{self.get_admin_preview_suffixes_default()}</a> '
            else:
                ret += f' <a class="cs_set_params" data-param-suffix="{suf}">{suf}</a> '
        ret += "</div>"

        return ret

    def give_python_to_admin(
        self, value: str, name: str, suffix: Optional[str] = None, **kwargs
    ) -> Any:
        return self.give_python(value, suffix)


class DictSuffixesPreviewMixin(DictSuffixesMixin, AdminPreviewSuffixesMixin):
    def get_admin_preview_suffixes(self, value: str, name: str, **kwargs):
        return tuple(self.suffixes.keys())



from django.conf import settings


def get_setting(name, default=None):
    return getattr(settings, "CONTENT_SETTINGS_" + name, default)


try:
    import pkg_resources

    VERSION = pkg_resources.get_distribution("django-content-settings").version
except Exception:
    VERSION = "dev"


CHECKSUM_KEY_PREFIX = (
    get_setting("CHECKSUM_KEY_PREFIX", "CS_CHECKSUM_") + VERSION + "__"
)

CHECKSUM_USER_KEY_PREFIX = (
    get_setting("CHECKSUM_KEY_PREFIX", "CS_USER_CHECKSUM_") + VERSION + "__"
)

USER_TAGS = get_setting(
    "USER_TAGS",
    {
        "favorites": ("⭐", "⚝"),
        "marked": ("💚", "♡"),
    },
)

CACHE_BACKEND = get_setting("CACHE_BACKEND", "default")

CACHE_TIMEOUT = get_setting("CACHE_TIMEOUT", 60 * 60 * 24)

CACHE_SPLITER = get_setting("CACHE_SPLITER", "::")

VALUES_ONLY_FROM_DB = get_setting("VALUES_ONLY_FROM_DB", False) and not settings.DEBUG

CONTEXT = get_setting("CONTEXT", {})

CONTEXT_PROCESSORS = get_setting("CONTEXT_PROCESSORS", [])

USER_DEFINED_TYPES = get_setting("USER_DEFINED_TYPES", [])

assert isinstance(USER_DEFINED_TYPES, list), "USER_DEFINED_TYPES must be a list"
assert len(USER_DEFINED_TYPES) == len(
    set([val[0] for val in USER_DEFINED_TYPES])
), "USER_DEFINED_TYPES must have unique slugs"
for i, val in enumerate(USER_DEFINED_TYPES):
    assert len(val) == 3, f"USER_DEFINED_TYPES[{i}] must be a tuple of 3 elements"
    assert all(
        [isinstance(v, str) for v in val]
    ), f"USER_DEFINED_TYPES[{i}] must be a tuple of 3 strings"

    slug, imp_line, name = val
    assert len(slug) < 50, f"USER_DEFINED_TYPES[{i}][0] must be less than 50 chars"



def any(user):
    return True


def none(user):
    return False


def authenticated(user):
    return user.is_authenticated


def staff(user):
    return user.is_active and user.is_staff


def superuser(user):
    return user.is_active and user.is_superuser


def has_perm(perm):
    def _(user):
        return user.is_active and user.has_perm(perm)

    return _


def and_(*funcs):
    def _(user):
        return all(func(user) for func in funcs)

    return _


def or_(*funcs):
    def _(user):
        return any(func(user) for func in funcs)

    return _


def not_(func):
    def _(user):
        return not func(user)

    return _



from datetime import datetime, time, timedelta
from django.core.exceptions import ValidationError
from django.utils import formats

from .basic import SimpleString, PREVIEW_PYTHON

TIMEDELTA_FORMATS = {
    "s": "seconds",
    "m": "minutes",
    "h": "hours",
    "d": "days",
    "w": "weeks",
}


def timedelta_format(text):
    text = text.strip()
    delta_kwargs = {}
    for el in text.split():
        try:
            val = int(el[:-1])
            val_key = el[-1]
            delta_kwargs[TIMEDELTA_FORMATS[val_key]] = val
        except Exception as e:
            raise ValidationError(f"wrong format for {el}: {e}")

    return timedelta(**delta_kwargs)


class DateTimeString(SimpleString):
    input_formats = formats.get_format_lazy("DATETIME_INPUT_FORMATS")
    empty_is_none = True
    admin_preview_as = PREVIEW_PYTHON

    def get_input_formats(self):
        if isinstance(self.input_formats, str):
            return [self.input_formats]
        return self.input_formats

    def get_help_format(self):
        yield (
            "Use any of the following formats: <ul>"
            + "".join([f"<li>{l}</li>" for l in self.get_input_formats()])
            + "</ul>"
        )

    def to_python(self, value):
        value = super().to_python(value)
        if value is None:
            return None
        value = value.strip()

        for format in self.get_input_formats():
            try:
                return self.postprocess_datetime(self.strptime(value, format))
            except (ValueError, TypeError):
                continue
        raise ValidationError("Wrong Format")

    def postprocess_datetime(self, value):
        from django.forms.utils import from_current_timezone

        return from_current_timezone(value)

    def strptime(self, value, format):
        return datetime.strptime(value, format)


class DateString(DateTimeString):
    input_formats = formats.get_format_lazy("DATE_INPUT_FORMATS")

    def postprocess_datetime(self, value):
        return value.date()


class TimeString(DateTimeString):
    input_formats = formats.get_format_lazy("TIME_INPUT_FORMATS")

    def postprocess_datetime(self, value):
        return time(
            value.hour, value.minute, value.second, value.microsecond, fold=value.fold
        )


class SimpleTimedelta(SimpleString):
    admin_preview_as = PREVIEW_PYTHON

    def get_help_format(self):
        yield "Time Delta Format"
        yield "<ul>"
        for name, value in TIMEDELTA_FORMATS.items():
            yield f"<li>{name} - {value}</li>"
        yield "</ul>"

        yield """
            Examples:
            <ul>
                <li>1d - in one day</li>
                <li>1d 3h - in one day and 3 hours</li>
            </ul>
        """

    def to_python(self, value):
        return timedelta_format(super().to_python(value))



from django.db.models.signals import post_save, post_migrate, post_delete, pre_save
from django.core.signals import request_started
from django.db.backends.signals import connection_created
from django.apps import apps
from django.db import transaction

from .settings import USER_DEFINED_TYPES

from django.dispatch import receiver

from .caching import (
    check_update,
    reset_all_values,
    recalc_checksums,
)
from .conf import set_initial_values_for_db
from .models import ContentSetting, HistoryContentSetting


@receiver(post_delete, sender=ContentSetting)
@receiver(post_save, sender=ContentSetting)
def do_update_stored_checksum(*args, **kwargs):
    connection = transaction.get_connection()
    if connection.in_atomic_block:
        transaction.on_commit(recalc_checksums)
    else:
        recalc_checksums()


@receiver(post_save, sender=ContentSetting)
def create_history_settings(sender, instance, created, **kwargs):
    HistoryContentSetting.objects.create(
        name=instance.name,
        value=instance.value,
        version=instance.version,
        tags=instance.tags,
        help=instance.help,
        user_defined_type=instance.user_defined_type,
        was_changed=not created,
    )


if USER_DEFINED_TYPES:

    @receiver(pre_save, sender=ContentSetting)
    def check_user_defined_type_version(sender, instance, **kwargs):
        from .conf import USER_DEFINED_TYPES_INITIAL

        if instance.user_defined_type not in USER_DEFINED_TYPES_INITIAL:
            return
        instance.version = USER_DEFINED_TYPES_INITIAL[
            instance.user_defined_type
        ].version


@receiver(post_delete, sender=ContentSetting)
def create_history_settings_delete(sender, instance, **kwargs):
    HistoryContentSetting.objects.create(
        name=instance.name,
        value=instance.value,
        version=instance.version,
        tags=instance.tags,
        help=instance.help,
        was_changed=None,
    )


@receiver(request_started)
def check_update_for_request(*args, **kwargs):
    check_update()


@receiver(post_migrate, sender=apps.get_app_config("content_settings"))
def update_stored_values(*args, **kwargs):
    try:
        with transaction.atomic():
            log = set_initial_values_for_db(apply=True)
    except Exception as e:
        log = [("Error", str(e))]

    if not log:
        return

    for l in log:
        print(" ".join(l))


@receiver(connection_created)
def db_connection_done(*args, **kwargs):
    reset_all_values()


# INTEGRATIONS

try:
    from celery.signals import task_prerun
except ImportError:
    pass
else:

    @task_prerun.connect
    def check_update_for_celery(*args, **kwargs):
        check_update()


try:
    from huey.contrib.djhuey import pre_execute
except ImportError:
    pass
else:

    @pre_execute()
    def check_update_for_huey(*args, **kwargs):
        check_update()



from threading import local
import hashlib
from functools import lru_cache

from django.core.cache import caches
from django.conf import settings

from .settings import (
    CHECKSUM_KEY_PREFIX,
    CHECKSUM_USER_KEY_PREFIX,
    CACHE_SPLITER,
    CACHE_TIMEOUT,
    CACHE_BACKEND,
    VALUES_ONLY_FROM_DB,
    USER_DEFINED_TYPES,
)


DATA = local()


def get_cache():
    return caches[CACHE_BACKEND]


@lru_cache(maxsize=None)
def get_cache_key():
    from .conf import ALL

    return calc_checksum({name: ALL[name].version for name in ALL.keys()})


@lru_cache(maxsize=None)
def get_user_cache_key():
    if not USER_DEFINED_TYPES:
        return None

    from .conf import USER_DEFINED_TYPES_INITIAL

    return calc_checksum({k: v.version for k, v in USER_DEFINED_TYPES_INITIAL.items()})


def calc_checksum(values):
    """
    generate md5 hash for a dict with keys and values as strings
    """
    return hash_value(
        CACHE_SPLITER.join(
            f"{name}{CACHE_SPLITER}{values[name]}" for name in sorted(values.keys())
        )
    )


def hash_value(value):
    """
    generate md5 hash for a string
    """
    return hashlib.md5(value.encode("utf-8")).hexdigest()


def set_local_checksum():
    from .conf import ALL

    DATA.ALL_VALUES_CHECKSUM = calc_checksum(
        {k: v for k, v in DATA.ALL_RAW_VALUES.items() if k in ALL}
    )


def set_local_user_checksum():
    from .conf import ALL

    DATA.ALL_VALUES_USER_CHECKSUM = calc_checksum(
        {k: v for k, v in DATA.ALL_RAW_VALUES.items() if k not in ALL}
    )


def push_user_checksum(value=None, key=None):
    if key is None:
        key = get_user_cache_key()

    if value is None:
        value = DATA.ALL_VALUES_USER_CHECKSUM

    get_cache().set(CHECKSUM_USER_KEY_PREFIX + key, value, CACHE_TIMEOUT)


def push_checksum(value=None, key=None):
    if key is None:
        key = get_cache_key()

    if value is None:
        value = DATA.ALL_VALUES_CHECKSUM

    get_cache().set(CHECKSUM_KEY_PREFIX + key, value, CACHE_TIMEOUT)


def get_checksum_from_store():
    return get_cache().get(CHECKSUM_KEY_PREFIX + get_cache_key())


def get_user_checksum_from_store():
    return get_cache().get(CHECKSUM_USER_KEY_PREFIX + get_user_cache_key())


def get_checksum_from_local():
    return DATA.ALL_VALUES_CHECKSUM


def get_checksum_from_user_local():
    return DATA.ALL_VALUES_USER_CHECKSUM


def set_new_type(name, cs):
    from .conf import USER_DEFINED_TYPES_INSTANCE

    cs_type = DATA.ALL_USER_DEFINES.get(name)

    if not cs_type or cs_type.tags != cs.tags_set or cs_type.help != cs.help:
        cs_type = USER_DEFINED_TYPES_INSTANCE[cs.user_defined_type](
            help=cs.help,
            tags=cs.tags_set,
        )

    DATA.ALL_USER_DEFINES[name] = cs_type


def set_new_value(name, new_value, version=None):
    cs_type = get_type_by_name(name)
    if cs_type is None:
        return None

    prev_value = DATA.ALL_RAW_VALUES.get(name)

    if version is None or cs_type.version == version and prev_value != new_value:
        DATA.ALL_RAW_VALUES[name] = new_value
        DATA.ALL_VALUES[name] = cs_type.to_python(new_value)

    return prev_value


def delete_user_value(name):
    if name not in DATA.ALL_RAW_VALUES:
        return None

    prev_value = DATA.ALL_RAW_VALUES.get(name)
    del DATA.ALL_RAW_VALUES[name]
    del DATA.ALL_VALUES[name]
    del DATA.ALL_USER_DEFINES[name]

    return prev_value


def get_type_by_name(name):
    from .conf import ALL

    if name in ALL:
        return ALL[name]

    if name in DATA.ALL_USER_DEFINES:
        return DATA.ALL_USER_DEFINES[name]

    return None


def get_value(name, suffix=None):
    assert DATA.POPULATED

    return get_type_by_name(name).give(DATA.ALL_VALUES[name], suffix)


def is_populated():  # better name?
    return getattr(DATA, "POPULATED", False)


def get_db_objects():
    from .models import ContentSetting

    return {v.name: v for v in ContentSetting.objects.all()}


def get_all_names():
    if not is_populated():
        return []
    return list(DATA.ALL_VALUES.keys())


def reset_all_values():
    if not is_populated():
        DATA.ALL_VALUES = {}
        DATA.ALL_RAW_VALUES = {}
        DATA.ALL_VALUES_CHECKSUM = None
        DATA.ALL_USER_DEFINES = {}
        DATA.POPULATED = False

    try:
        db = get_db_objects()
    except Exception:
        DATA.POPULATED = False
        return
    if USER_DEFINED_TYPES:
        reset_user_values(db)
    reset_code_values(db)

    DATA.POPULATED = True


def reset_user_values(db=None, trigger_checksum=None):

    from .conf import USER_DEFINED_TYPES_INSTANCE

    if db is None:
        db = get_db_objects()

    names = set()
    for name, cs in db.items():
        if not cs.user_defined_type:
            continue

        assert (
            cs.user_defined_type in USER_DEFINED_TYPES_INSTANCE
        ), f"{cs.user_defined_type} is not in USER_DEFINED_TYPES"
        names.add(name)
        set_new_type(
            name,
            cs,
        )
        set_new_value(
            name,
            db[name].value,
            version=(None if settings.DEBUG else db[name].version),
        )

    for name in set(DATA.ALL_USER_DEFINES.keys()) - names:
        delete_user_value(name)

    set_local_user_checksum()

    if trigger_checksum != get_checksum_from_user_local():
        push_user_checksum()


def reset_code_values(db=None, trigger_checksum=None):
    from .conf import ALL

    if db is None:
        db = get_db_objects()

    for name, cs_type in ALL.items():
        if cs_type.constant:
            set_new_value(name, ALL[name].default)

        elif name in db:
            assert (
                not db[name].user_defined_type or cs_type.overwrite_user_defined
            ), f"{name} is not a code setting and not overwrite_user_defined"

            set_new_value(
                name,
                db[name].value,
                version=(None if settings.DEBUG else db[name].version),
            )

        else:
            if VALUES_ONLY_FROM_DB:
                raise AssertionError(f"VALUES_ONLY_FROM_DB: {name} is not in DB")
            set_new_value(name, ALL[name].default, version=ALL[name].version)

    set_local_checksum()

    if trigger_checksum != get_checksum_from_local():
        push_checksum()


def check_update():
    if not is_populated():
        return
    store_checksum = get_checksum_from_store()
    if store_checksum is not None and store_checksum != get_checksum_from_local():
        reset_code_values(trigger_checksum=store_checksum)

    if USER_DEFINED_TYPES:
        store_user_checksum = get_user_checksum_from_store()
        if (
            store_user_checksum is not None
            and store_user_checksum != get_checksum_from_user_local()
        ):
            reset_user_values(trigger_checksum=store_user_checksum)


def recalc_checksums():
    db = get_db_objects()
    recalc_code_checksums(db)
    if USER_DEFINED_TYPES:
        recalc_user_checksums(db)


def recalc_user_checksums(db):
    from .conf import USER_DEFINED_TYPES_INITIAL

    values = {}
    for name, cs in db.items():
        if not cs.user_defined_type:
            continue
        if cs.user_defined_type not in USER_DEFINED_TYPES_INITIAL:
            break
        if cs.version == USER_DEFINED_TYPES_INITIAL[cs.user_defined_type].version:
            values[name] = cs.user_defined_type + CACHE_SPLITER + db[name].value
        elif name in DATA.ALL_RAW_VALUES:
            values[name] = (
                cs.user_defined_type + CACHE_SPLITER + DATA.ALL_RAW_VALUES[name]
            )
        else:
            break
    else:
        push_user_checksum(calc_checksum(values))


def recalc_code_checksums(db):
    from .conf import ALL

    push_checksum(
        calc_checksum({k: v.value for k, v in db.items()}),
        calc_checksum({k: v.version for k, v in db.items()}),
    )

    db_versions = {}
    db_values = {}

    for name in set(ALL.keys()) | set(db.keys()):
        if name not in ALL or ALL[name].constant:
            continue

        if name in db:
            db_versions[name] = db[name].version
            db_values[name] = db[name].value
        else:
            db_versions[name] = ALL[name].version
            db_values[name] = ALL[name].default

    db_version_key = calc_checksum(db_versions)
    push_checksum(calc_checksum(db_values), db_version_key)

    if db_version_key == get_cache_key() or not is_populated():
        return

    local_values = {}
    for name in ALL.keys():
        if name in db:
            db_version = db[name].version
            if db_version != ALL[name].version:
                local_values[name] = DATA.ALL_RAW_VALUES[name]
            else:
                local_values[name] = db[name].value
        else:
            local_values[name] = ALL[name].default

    push_checksum(calc_checksum(local_values))



from contextlib import contextmanager, ContextDecorator
from .settings import CONTEXT, CONTEXT_PROCESSORS

CONTEXT_DEFAULTS_STACK = [CONTEXT]
CONTEXT_DEFAULTS_PROCESSORS_STACK = [CONTEXT_PROCESSORS]


class process:
    """
    the flag _init determines whether the processor should be applied for default values
    of initial (processed after default values)
    """

    init = False

    def __init__(self, **kwargs):
        if "_init" in kwargs:
            self.init = kwargs.pop("_init")
        self.kwargs = kwargs

    def process(self, kwargs):
        raise NotImplementedError()


class process_set(process):
    def process(self, kwargs):
        return {
            name: self.process_one(
                kwargs.get(name, set()),
                set([value]) if isinstance(value, str) else set(value),
            )
            for name, value in self.kwargs.items()
        }


class set_add(process_set):
    def process_one(self, start, change):
        return start | change


class set_remove(process_set):
    def process_one(self, start, change):
        return start - change


class add_tags(set_add):
    def __init__(self, tags):
        super().__init__(tags=tags, _init=True)


class remove_tags(set_remove):
    def __init__(self, tags):
        super().__init__(tags=tags, _init=True)


class process_str(process):
    def process(self, kwargs):
        return {
            name: self.process_one(
                kwargs.get(name, ""),
                "" if value is None else str(value),
            )
            for name, value in self.kwargs.items()
        }


class str_append(process_str):
    def process_one(self, start, change):
        return start + change


class str_prepend(process_str):
    def process_one(self, start, change):
        return change + start


class str_format(process_str):
    def process_one(self, start, change):
        return change.format(start)


class help_format(str_format):
    def __init__(self, text):
        super().__init__(help=text, _init=True)


@contextmanager
def context_defaults(*args, **kwargs):
    CONTEXT_DEFAULTS_STACK.append(kwargs)
    CONTEXT_DEFAULTS_PROCESSORS_STACK.append(args)
    try:
        yield
    finally:
        CONTEXT_DEFAULTS_STACK.pop()
        CONTEXT_DEFAULTS_PROCESSORS_STACK.pop()


def context_defaults_kwargs(kwargs=None):
    init = True
    if kwargs is None:
        kwargs = {}
        init = False

    for processors, new_kwargs in zip(
        CONTEXT_DEFAULTS_PROCESSORS_STACK, CONTEXT_DEFAULTS_STACK
    ):
        for processor in processors:
            if processor.init != init:
                continue
            for name, value in processor.process(kwargs).items():
                kwargs[name] = value

        if not init:
            for name, value in new_kwargs.items():
                kwargs[name] = value

    return kwargs


class content_settings_context(ContextDecorator):
    def __init__(self, **values) -> None:
        super().__init__()
        self.values_to_update = values
        self.prev_values = {}

    def __enter__(self):
        from content_settings.caching import set_new_value

        for name, new_value in self.values_to_update.items():
            self.prev_values[name] = set_new_value(name, new_value)

    def __exit__(self, *exc):
        from content_settings.caching import set_new_value

        for name, new_value in self.prev_values.items():
            set_new_value(name, new_value)



from collections import defaultdict

from django.db import models
from django.core.validators import RegexValidator
from django.utils import timezone
from django.conf import settings


class ContentSetting(models.Model):
    name = models.CharField(
        max_length=200,
        unique=True,
        validators=[
            RegexValidator(
                regex=r"^[A-Z][A-Z0-9_]*$",
                message="Value must be uppercase, can include numbers and underscores, and cannot start with a number.",
            ),
        ],
    )
    value = models.TextField(blank=True)
    version = models.CharField(max_length=50, null=True)
    help = models.TextField(blank=True, null=True)
    tags = models.TextField(blank=True, null=True)
    user_defined_type = models.CharField(
        max_length=50, null=True, default=None, db_index=True
    )

    class Meta:
        ordering = ("name",)

    @property
    def tags_set(self):
        if not self.tags:
            return set()

        return set(tag for tag in self.tags.splitlines() if tag.strip())

    def __str__(self):
        return self.name


class HistoryContentSetting(models.Model):
    CHANGED_CHOICES = (
        (True, "Changed"),
        (False, "Added"),
        (None, "Removed"),
    )
    BY_USER_CHOICES = (
        (True, "by user"),
        (False, "by app"),
        (None, "unknown"),
    )

    created_on = models.DateTimeField(null=False, default=timezone.now)

    name = models.CharField(max_length=200)
    value = models.TextField(blank=True, null=True)
    version = models.CharField(max_length=50, null=True)

    was_changed = models.BooleanField(null=True, default=True, choices=CHANGED_CHOICES)
    by_user = models.BooleanField(null=True, default=None, choices=BY_USER_CHOICES)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.SET_NULL, null=True
    )
    help = models.TextField(blank=True, null=True)
    tags = models.TextField(blank=True, null=True)
    user_defined_type = models.CharField(max_length=50, null=True, default=None)

    @property
    def user_defined_type_display(self):
        from .conf import USER_DEFINED_TYPES_NAME

        if not self.user_defined_type:
            return "---"

        if self.user_defined_type in USER_DEFINED_TYPES_NAME:
            return USER_DEFINED_TYPES_NAME[self.user_defined_type]

        return f"{self.user_defined_type} (unknown)"

    class Meta:
        ordering = ("-id",)
        indexes = [
            models.Index(fields=["name", "-id"]),
        ]

    @classmethod
    def update_last_record_for_name(cls, name, user=None):
        last_setting = cls.objects.filter(name=name).first()
        last_setting.user = user
        last_setting.by_user = user is not None
        last_setting.save()

    @classmethod
    def gen_unique_records(cls, name):
        next_obj = None
        records = cls.objects.filter(name=name)
        if not records.exists():
            return []

        for obj in records:
            if next_obj is None:
                next_obj = obj
                continue

            if (
                next_obj.value != obj.value
                or next_obj.was_changed != obj.was_changed
                or next_obj.tags != obj.tags
                or next_obj.help != obj.help
                or next_obj.user_defined_type != obj.user_defined_type
            ):
                yield next_obj

            next_obj = obj

        yield obj


class UserTagSetting(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    name = models.CharField(max_length=200)
    tag = models.TextField()

    class Meta:
        unique_together = (("user", "name", "tag"),)

    def __str__(self):
        return f"{self.user} - {self.name} - {self.tag}"

    @classmethod
    def get_user_settings(cls, user):
        settings = defaultdict(set)
        for name, tag in cls.objects.filter(user=user).values_list("name", "tag"):
            settings[name].add(tag)
        return settings



# Generated by Django 2.2.12 on 2020-06-01 19:35

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="ContentSetting",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=200, unique=True)),
                ("value", models.TextField(blank=True)),
                ("version", models.CharField(max_length=50, null=True)),
            ],
            options={"ordering": ("name",)},
        ),
        migrations.CreateModel(
            name="HistoryContentSetting",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_on", models.DateTimeField(default=django.utils.timezone.now)),
                ("name", models.CharField(max_length=200)),
                ("value", models.TextField(blank=True, null=True)),
                ("version", models.CharField(max_length=50, null=True)),
                (
                    "was_changed",
                    models.BooleanField(
                        choices=[
                            (True, "Changed"),
                            (False, "Added"),
                            (None, "Removed"),
                        ],
                        default=True,
                        null=True,
                    ),
                ),
                (
                    "by_user",
                    models.BooleanField(
                        choices=[
                            (True, "by user"),
                            (False, "by app"),
                            (None, "unknown"),
                        ],
                        default=None,
                        null=True,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "ordering": ("-id",),
            },
        ),
        migrations.CreateModel(
            name="UserTagSetting",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=200)),
                ("tag", models.TextField()),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "unique_together": {("user", "name", "tag")},
            },
        ),
        migrations.AddIndex(
            model_name="historycontentsetting",
            index=models.Index(
                fields=["name", "-id"], name="static_cont_name_d40020_idx"
            ),
        ),
    ]



from typing import Any, Optional

from django.core.exceptions import ValidationError

from .basic import SimpleText, BaseSetting, PREVIEW_PYTHON, PREVIEW_TEXT, PREVIEW_NONE
from .mixins import AdminPreviewSuffixesMixin


def f_empty(value):
    value = value.strip()
    if not value:
        return None
    return value


def f_comment(prefix):
    def _(value):
        if value.strip().startswith("#"):
            return None
        return value

    return _


class SimpleStringsList(SimpleText):
    comment_starts_with = "#"
    filter_empty = True
    split_lines = "\n"
    filters = None
    admin_preview_as = PREVIEW_PYTHON

    def get_filters(self):
        if self.filters is not None:
            return self.filters

        filters = []
        if self.filter_empty:
            filters.append(f_empty)

        if self.comment_starts_with:
            filters.append(f_comment(self.comment_starts_with))

        return filters

    def get_help_format(self):
        yield "List of values with the following format:"
        yield "<ul>"
        if self.split_lines == "\n":
            yield "<li>each line is a new value</li>"
        else:
            yield f"<li>{self.split_lines} separates values</li>"
        yield "<li>strip spaces from the beginning and from the end of the value</li>"
        yield "<li>remove empty values</li>"
        if self.comment_starts_with:
            yield f"<li> use {self.comment_starts_with} to comment a line</li>"
        if self.filter_empty:
            yield "<li>empty values are removed</li>"
        yield "</ul>"

    def filter_line(self, line):
        for f in self.get_filters():
            line = f(line)
            if line is None:
                break
        return line

    def gen_to_python(self, value):
        lines = value.split(self.split_lines)
        for line in lines:
            line = self.filter_line(line)

            if line is not None:
                yield line

    def to_python(self, value):
        return list(self.gen_to_python(value))


class TypedStringsList(SimpleStringsList):
    line_type = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        assert self.line_type is not None, "line_type should be set"
        assert isinstance(
            self.line_type, BaseSetting
        ), "line_type should be BaseSetting"

    def validate_raw_value(self, value):
        lines = value.split(self.split_lines)
        for num, line in enumerate(lines, 1):
            line = self.filter_line(line)
            if line is None:
                continue

            try:
                self.line_type.validate_value(line)
            except ValidationError as e:
                raise ValidationError(f"Line {num}: {e.message}")

        return super().validate_value(value)

    def gen_to_python(self, value):
        for line in super().gen_to_python(value):
            yield self.line_type.to_python(line)

    def get_help_format(self):
        yield from super().get_help_format()
        yield "Each line is "
        yield from self.line_type.get_help_format()

    def give(self, value, suffix=None):
        return [self.line_type.give(v) for v in value]


NOT_FOUND_DEFAULT = "default"
NOT_FOUND_KEY_ERROR = "key_error"
NOT_FOUND_VALUE = "value"

SPLIT_SUFFIX_USE_OWN = "own"
SPLIT_SUFFIX_USE_PARENT = "parent"
SPLIT_SUFFIX_SPLIT_OWN = "split_own"
SPLIT_SUFFIX_SPLIT_PARENT = "split_parent"


class SplitByFirstLine(AdminPreviewSuffixesMixin, SimpleText):
    split_type = SimpleText()
    split_values = None
    split_default_key = None
    split_default_chooser = None
    split_not_found = NOT_FOUND_DEFAULT
    split_not_found_value = None
    split_key_validator = None
    split_suffix = SPLIT_SUFFIX_USE_OWN
    split_suffix_value = None

    def get_split_default_key(self):
        return self.split_default_key

    def split_default_choose(self, value):
        if self.split_default_chooser is None:
            return self.get_split_default_key()
        else:
            return self.split_default_chooser(value)

    def get_split_not_found(self):
        return self.split_not_found

    def get_split_not_found_value(self):
        return self.split_not_found_value

    def get_split_key_validator(self):
        return self.split_key_validator

    def get_split_suffix(self):
        return self.split_suffix

    def get_split_suffix_value(self):
        return self.split_suffix_value

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        assert (
            self.get_split_default_key()
        ), "split_default_key should be set and it should not be empty"
        assert isinstance(
            self.get_split_default_key(), str
        ), "split_default_key should be str"

        assert self.split_default_chooser is None or callable(
            self.split_default_chooser
        ), "split_default_chooser should be callable or None"
        assert self.get_split_key_validator() is None or callable(
            self.get_split_key_validator()
        ), "split_key_validator should be callable or None"
        assert self.get_split_not_found() in (
            NOT_FOUND_DEFAULT,
            NOT_FOUND_KEY_ERROR,
            NOT_FOUND_VALUE,
        ), "split_not_found should be one of NOT_FOUND_DEFAULT, NOT_FOUND_KEY_ERROR, NOT_FOUND_VALUE"
        assert self.get_split_suffix_value() is not None or self.get_split_suffix() in (
            SPLIT_SUFFIX_USE_OWN,
            SPLIT_SUFFIX_USE_PARENT,
        ), "split_suffix should be one of SPLIT_SUFFIX_USE_OWN, SPLIT_SUFFIX_SPLIT_PARENT or split_suffix_value should be set"
        assert self.get_split_suffix_value() is None or isinstance(
            self.get_split_suffix_value(), str
        ), "split_suffix_value should be str or None"
        assert self.get_split_suffix() in (
            SPLIT_SUFFIX_USE_OWN,
            SPLIT_SUFFIX_USE_PARENT,
            SPLIT_SUFFIX_SPLIT_OWN,
            SPLIT_SUFFIX_SPLIT_PARENT,
        ), "split_suffix should be one of SPLIT_SUFFIX_USE_OWN, SPLIT_SUFFIX_USE_PARENT, SPLIT_SUFFIX_SPLIT_OWN, SPLIT_SUFFIX_SPLIT_PARENT, SPLIT_SUFFIX_SPLIT_FUNCTION"
        assert self.split_values is None or callable(
            self.split_values
        ), "split_values should be callable or None"

    def split_value(self, value):
        if self.split_values is not None:
            return self.split_values(value)
        lines = value.split("\n")
        if (
            self.get_split_default_key() not in lines[0]
            or lines[0].count(self.get_split_default_key()) > 1
        ):
            return {self.get_split_default_key(): value}

        before, after = lines[0].split(self.get_split_default_key())
        ret = {self.get_split_default_key(): []}
        cur_iter = ret[self.get_split_default_key()]
        for line in lines[1:]:
            if line.startswith(before) and line.endswith(after):
                cur_iter = ret[line[len(before) : len(line) - len(after)]] = []
                continue

            cur_iter.append(line)

        return {k: "\n".join(v) for k, v in ret.items()}

    def get_admin_preview_as(self):
        split_admin_preview = self.split_type.get_admin_preview_as()
        if split_admin_preview == PREVIEW_NONE:
            return PREVIEW_TEXT
        return split_admin_preview

    def get_admin_preview_suffixes_default(self):
        return self.get_split_default_key()

    def get_admin_preview_suffixes(self, value: str, name: str, **kwargs):
        keys = list(self.split_value(value).keys())
        keys.remove(self.get_split_default_key())
        return tuple(keys)

    def to_python(self, value):
        values = self.split_value(value)
        ret = {}
        for k, v in values.items():
            try:
                ret[k] = self.split_type.to_python(v)
            except ValidationError as e:
                if k == self.get_split_default_key():
                    raise
                raise ValidationError(f"{k}: {e}")
        return ret

    def give_siffixes(self, value):
        if value is None:
            return None, None

        if self.get_split_suffix() == SPLIT_SUFFIX_USE_OWN:
            return value.upper(), None

        if self.get_split_suffix() == SPLIT_SUFFIX_USE_PARENT:
            return None, value

        if self.get_split_suffix_value() not in value:
            if self.get_split_suffix() == SPLIT_SUFFIX_SPLIT_OWN:
                return value.upper(), None
            else:
                return None, value

        suffix, parent_suffix = value.split(self.get_split_suffix_value(), 1)
        suffix = suffix.upper()
        return suffix, parent_suffix

    def give(self, value, suffix=None):
        key, parent_suffix = self.give_siffixes(suffix)
        if key is None:
            key = self.split_default_choose(value)

        try:
            ret_val = value[key]
        except KeyError:
            if self.get_split_not_found() == NOT_FOUND_KEY_ERROR:
                raise
            elif self.get_split_not_found() == NOT_FOUND_DEFAULT:
                ret_val = value[self.get_split_default_key()]
            elif self.get_split_not_found() == NOT_FOUND_VALUE:
                ret_val = self.get_split_not_found_value()

        return self.split_type.give(ret_val, parent_suffix)

    def validate_raw_value(self, value):
        values = self.split_value(value)

        for k, v in values.items():
            try:
                self.split_type.validate_value(v)
            except Exception as e:
                if k == self.get_split_default_key():
                    raise
                raise ValidationError(f"{k}: {e}")

    def validate(self, values):
        if self.get_split_key_validator() is not None:
            for k in values.keys():
                self.get_split_key_validator()(k)

    def get_help_format(self):
        yield from super().get_help_format()
        yield from self.split_type.get_help_format()


class SplitTranslation(SplitByFirstLine):
    split_default_key = "EN"
    split_not_found = NOT_FOUND_DEFAULT

    def get_help_format(self):
        yield f"<b>With translation ({self.get_split_default_key()} by default)</b>"

        yield from super().get_help_format()
        yield from self.split_type.get_help_format()

    def split_default_choose(self, value):
        from django.utils.translation import get_language

        return get_language().upper().replace("-", "_")



from importlib import import_module
from functools import partial

from django.apps import apps

from .types.basic import BaseSetting
from .caching import get_value, get_type_by_name, get_all_names
from .settings import USER_DEFINED_TYPES

USER_DEFINED_TYPES_INSTANCE = {}
USER_DEFINED_TYPES_INITIAL = {}
USER_DEFINED_TYPES_NAME = {}
ALL = {}
PREFIXSES = {}

if USER_DEFINED_TYPES:
    for slug, imp_line, name in USER_DEFINED_TYPES:
        imp_line_parts = imp_line.split(".")
        type_module = import_module(".".join(imp_line_parts[:-1]))
        type_class = getattr(type_module, imp_line_parts[-1])
        USER_DEFINED_TYPES_INSTANCE[slug] = partial(
            type_class, user_defined_slug=slug, version=type_class.version
        )
        USER_DEFINED_TYPES_INITIAL[slug] = USER_DEFINED_TYPES_INSTANCE[slug]()
        USER_DEFINED_TYPES_NAME[slug] = name


def register_prefix(name):
    def _cover(func):
        PREFIXSES[name] = func
        return func

    return _cover


@register_prefix("lazy")
def lazy_prefix(name, suffix):
    return get_type_by_name(name).lazy_give(lambda: get_value(name, suffix), suffix)


@register_prefix("type")
def type_prefix(name, suffix):
    assert not suffix, "type prefix can not have suffix"

    return get_type_by_name(name)


@register_prefix("startswith")
def startswith_prefix(name, suffix):
    return content_settings.startswith(name, suffix)


@register_prefix("withtag")
def withtag_prefix(name, suffix):
    return {
        **content_settings.withtag(name, suffix),
        **content_settings.withtag(name.lower(), suffix),
    }


for app_config in apps.app_configs.values():
    app = app_config.name
    try:
        content_settings = import_module(app + ".content_settings")
    except ImportError as e:
        if e.name != app + ".content_settings":
            raise
        continue
    for attr in dir(content_settings):
        val = getattr(content_settings, attr)
        if not isinstance(val, BaseSetting):
            continue

        assert attr not in ALL, "Content Setting {} defined twice".format(attr)

        if attr in ALL and not ALL[attr].user_defined_slug:
            raise AssertionError("Overwriting content setting {}".format(attr))

        if not attr.isupper():
            raise AssertionError("content setting {} should be uppercase".format(attr))

        assert (
            not val.user_defined_slug
        ), "Do not set user_defined_slug in content_settings.py"

        ALL[attr] = val


def split_attr(value):
    """
    splits the name of the attr on 3 parts: prefix, name, suffix

    * prefix should be registered by register_prefix
    * name should be uppercase
    * suffix can be any string, but not uppercase
    """
    prefix = None
    parts = value.split("__")

    if parts[0] in PREFIXSES:
        prefix = parts.pop(0)

    assert len(parts), f"Invalid attribute name: {value}; can not be only prefix"

    name = parts.pop(0)
    assert name.isupper(), f"Invalid attribute name: {value}; name should be uppercase"

    while parts:
        if not parts[0].isupper():
            break

        name += "__" + parts.pop(0)

    if len(parts):
        return prefix, name, "__".join(parts).lower()

    return prefix, name, None


class _Settings:
    def __getattr__(self, value):
        prefix, name, suffix = split_attr(value)
        if prefix:
            assert (
                prefix in PREFIXSES
            ), f"Invalid attribute name: {value}; prefix not found"
            return PREFIXSES[prefix](name, suffix)
        return get_value(name, suffix)

    def __dir__(self):
        return get_all_names()

    def startswith(self, value, suffix=None):
        return {k: get_value(k, suffix) for k in dir(self) if k.startswith(value)}

    def withtag(self, value, suffix=None):
        return {
            k: get_value(k, suffix)
            for k in dir(self)
            if value in get_type_by_name(k).get_tags()
        }

    def __contains__(self, value):
        _, name, suffix = split_attr(value)
        cs_type = get_type_by_name(name)
        return cs_type is not None and cs_type.can_suffix(suffix)


def get_str_tags(cs_type):
    if cs_type.get_tags():
        return "\n".join(sorted(cs_type.get_tags()))
    return ""


def set_initial_values_for_db(apply=False):
    from content_settings.models import ContentSetting, HistoryContentSetting

    changes = []

    def execute(name, key, func):
        changes.append((name, key))
        if apply:
            func()
            HistoryContentSetting.update_last_record_for_name(name)

    def execute_update_obj(name, cs, **kwargs):
        def _up():
            for k, v in kwargs.items():
                setattr(cs, k, v)
            cs.save()

        execute(name, "update", _up)

    for k, cs_type in ALL.items():
        if cs_type.constant:
            continue

        try:
            ContentSetting.objects.get(name=k)
        except ContentSetting.DoesNotExist:
            execute(
                k,
                "create",
                lambda: ContentSetting.objects.create(
                    name=k,
                    value=cs_type.default,
                    version=cs_type.version,
                    tags=get_str_tags(cs_type),
                    help=cs_type.get_help(),
                ),
            )

    for cs in ContentSetting.objects.all():
        if cs.name in ALL:
            cs_type = ALL[cs.name]
            if cs_type.constant:
                execute(cs.name, "delete", lambda: cs.delete())
                continue

            str_tags = get_str_tags(cs_type)
            str_help = cs_type.get_help()

            assert (
                not cs.user_defined_type or cs_type.overwrite_user_defined
            ), f"{cs.name} is not a code setting and not overwrite_user_defined"

            if (
                cs.version != cs_type.version
                or cs.user_defined_type
                or cs.tags != str_tags
                or cs.help != str_help
            ):
                execute_update_obj(
                    cs.name,
                    cs,
                    value=cs_type.default,
                    version=cs_type.version,
                    tags=str_tags,
                    help=str_help,
                    user_defined_type=None,
                )

        else:
            if cs.user_defined_type:
                if cs.user_defined_type not in USER_DEFINED_TYPES_INSTANCE:
                    execute(cs.name, "delete", lambda: cs.delete())
                elif (
                    cs.version
                    != USER_DEFINED_TYPES_INITIAL[cs.user_defined_type].version
                ):
                    cs_type = USER_DEFINED_TYPES_INSTANCE[cs.user_defined_type]()
                    execute_update_obj(
                        cs.name, cs, value=cs_type.default, version=cs_type.version
                    )
            else:
                execute(cs.name, "delete", lambda: cs.delete())

    return changes


content_settings = _Settings()



from django.apps import AppConfig


class StaticContentConfig(AppConfig):
    name = "content_settings"
    verbose_name = "Content Settings"

    def ready(self):
        import content_settings.signals
        from content_settings.caching import reset_all_values

        reset_all_values()



import pytest


pytestmark = [
    pytest.mark.django_db,
]

from content_settings.conf import content_settings, ALL
from content_settings.caching import reset_all_values


@pytest.mark.parametrize(
    "name", [name for name in ALL.keys() if not ALL[name].constant]
)
def test_admin_fields(name):
    reset_all_values()
    value = getattr(content_settings, name)
    setting = ALL[name]
    raw_value = setting.default
    assert setting.get_admin_preview_value(raw_value, name) is not None
    assert setting.get_help()


@pytest.mark.parametrize("name", ALL.keys())
def test_validate_default(name):
    reset_all_values()
    setting = ALL[name]
    raw_value = setting.default
    setting.validate_value(raw_value)



from functools import cached_property

from django.core.exceptions import ValidationError

from .basic import SimpleText, PREVIEW_PYTHON, SimpleString


class SimpleYAML(SimpleText):
    help_format = "Simple <a href='https://en.wikipedia.org/wiki/YAML' target='_blank'>YAML format</a>"
    admin_preview_as = PREVIEW_PYTHON
    yaml_loader = None
    tags = {"yaml"}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        try:
            import yaml
        except ImportError:
            raise AssertionError("Please install pyyaml package")

    def get_yaml_loader(self):
        if self.yaml_loader is None:
            try:
                from yaml import CLoader as Loader
            except ImportError:
                from yaml import Loader
            return Loader
        return self.yaml_loader

    def to_python(self, value):
        value = super().to_python(value)
        if value is None:
            return None

        from yaml import load

        try:
            return load(value, Loader=self.get_yaml_loader())
        except Exception as e:
            raise ValidationError(str(e))


class SimpleJSON(SimpleText):
    help_format = "Simple <a href='https://en.wikipedia.org/wiki/JSON' target='_blank'>JSON format</a>"
    admin_preview_as = PREVIEW_PYTHON
    decoder_cls = None
    tags = {"json"}

    def get_decoder_cls(self):
        return self.decoder_cls

    def to_python(self, value):
        value = super().to_python(value)
        if value is None:
            return None
        from json import loads

        try:
            return loads(value, cls=self.get_decoder_cls())
        except Exception as e:
            raise ValidationError(str(e))


class SimpleCSV(SimpleText):
    help_format = "Simple <a href='https://en.wikipedia.org/wiki/Comma-separated_values' target='_blank'>CSV format</a>"
    admin_preview_as = PREVIEW_PYTHON
    csv_dialect = "unix"
    fields = None
    tags = {"csv"}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert self.fields is not None, "fields cannot be None"
        assert isinstance(
            self.fields, (list, tuple, dict)
        ), "fields must be list, tuple or dict"

    @cached_property
    def dict_fields(self):
        if isinstance(self.fields, dict):
            return self.fields
        return {f: SimpleString() for f in self.fields}

    def gen_to_python(self, value):
        value = super().to_python(value)
        if value is None:
            return None

        from csv import reader
        from io import StringIO

        try:
            csv = reader(StringIO(value), dialect=self.csv_dialect)
        except Exception as e:
            raise ValidationError(str(e))

        yield from self.gen_rows_to_python(csv)

    def gen_rows_to_python(self, csv):
        for row in csv:
            if not row:
                continue
            yield {
                name: c_type.to_python(value)
                for (name, c_type), value in zip(self.dict_fields.items(), row)
            }

    def to_python(self, value):
        val = self.gen_to_python(value)
        if val is None:
            return None
        return list(val)

    def give(self, value, suffix=None):
        if not value:
            return value
        return [
            {k: self.dict_fields[k].give(v) for k, v in row.items()} for row in value
        ]



import operator


class LazyObject:
    def __init__(self, factory):
        # Assign using __dict__ to avoid the setattr method.
        self.__dict__["_factory"] = factory

    def new_method_proxy(func):
        """
        Util function to help us route functions
        to the nested object.
        """

        def inner(self, *args):
            return func(self._factory(), *args)

        return inner

    def __call__(self, *args, **kwargs):
        return self._factory()(*args, **kwargs)

    __getattr__ = new_method_proxy(getattr)
    __bytes__ = new_method_proxy(bytes)
    __str__ = new_method_proxy(str)
    __bool__ = new_method_proxy(bool)
    __dir__ = new_method_proxy(dir)
    __hash__ = new_method_proxy(hash)
    __class__ = property(new_method_proxy(operator.attrgetter("__class__")))
    __eq__ = new_method_proxy(operator.eq)
    __lt__ = new_method_proxy(operator.lt)
    __le__ = new_method_proxy(operator.le)
    __gt__ = new_method_proxy(operator.gt)
    __ge__ = new_method_proxy(operator.ge)
    __ne__ = new_method_proxy(operator.ne)
    __mod__ = new_method_proxy(operator.mod)
    __hash__ = new_method_proxy(hash)
    __getitem__ = new_method_proxy(operator.getitem)
    __iter__ = new_method_proxy(iter)
    __len__ = new_method_proxy(len)
    __contains__ = new_method_proxy(operator.contains)



from django.http import HttpResponseNotFound, HttpResponseForbidden, JsonResponse
from django.views.generic import View
from content_settings.conf import content_settings

from .conf import ALL, split_attr, content_settings
from .caching import get_type_by_name


def gen_startswith(startswith):
    def _(request):
        for name in dir(content_settings):
            if name.startswith(startswith):
                yield name

    return _


def gen_hastag(tag):
    def _(request):
        for name in dir(content_settings):
            if tag in get_type_by_name(name).get_tags():
                yield name

    return _


class FetchSettingsView(View):
    attrs = ()

    def get(self, request):
        ret = {}
        for val in self.attrs(request) if callable(self.attrs) else self.attrs:
            prefix, name, suffix = split_attr(val)
            assert prefix is None, "prefix is not None"

            cs_type = get_type_by_name(name)
            if cs_type is None:
                continue

            if not cs_type.fetch_permission or not cs_type.fetch_permission(
                request.user
            ):
                continue
            if suffix is None:
                setting_name = name
            else:
                setting_name = f"{name}__{suffix}"
            value = getattr(content_settings, setting_name)
            ret[setting_name] = cs_type.json_view_value(request, value)

        return JsonResponse(ret)


def fetch_one_setting(request, name, suffix=None):
    settings_name = name.upper().replace("-", "_")
    cs_type = get_type_by_name(settings_name)
    if cs_type is None:
        return HttpResponseNotFound("value not found")

    if cs_type.fetch_permission is None:
        return HttpResponseNotFound("hidden")

    if not cs_type.fetch_permission(request.user):
        return HttpResponseForbidden("permission denied")

    if suffix is not None:
        suffix = suffix.lower().replace("-", "_")
        settings_name = f"{settings_name}__{suffix}"

    value = getattr(content_settings, settings_name)

    return JsonResponse({settings_name: cs_type.json_view_value(request, value)})



import urllib.parse
from collections import defaultdict

from django.contrib import admin
from django.forms import ModelForm
from django import forms
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.contrib.admin.views.main import ChangeList
from django.urls import path
from django.http import JsonResponse
from django.contrib.admin.utils import unquote
from django.core.exceptions import PermissionDenied
from django.utils.text import capfirst
from django.template.response import TemplateResponse
from django.contrib.messages import add_message, ERROR
from django.utils.translation import gettext as _
from django.db.models import Q


from .models import ContentSetting, HistoryContentSetting, UserTagSetting
from .context_managers import content_settings_context
from .conf import ALL, USER_DEFINED_TYPES_INSTANCE
from .settings import USER_TAGS, USER_DEFINED_TYPES
from .caching import get_type_by_name


def user_able_to_update(user, name, user_defined_type=None):
    cs_type = get_type_by_name(name)
    if cs_type is None and user_defined_type is not None:
        cs_type = USER_DEFINED_TYPES_INSTANCE[user_defined_type]()
    if cs_type is None:
        return False
    return cs_type.update_permission is None or cs_type.update_permission(user)


TAGS_PARAM = "tags"
TAGS_SPLITER = "|"


def get_selected_tags_from_params(params):
    if TAGS_PARAM not in params:
        return set()
    return set(params[TAGS_PARAM].split(TAGS_SPLITER))


class SettingsChangeList(ChangeList):
    def get_filters_params(self, *args, **kwargs):
        params = super().get_filters_params(*args, **kwargs)
        if TAGS_PARAM in params:
            del params[TAGS_PARAM]
        return params

    def get_queryset(self, request, *args, **kwargs):
        q = super().get_queryset(request, *args, **kwargs)

        tags = get_selected_tags_from_params(self.params)
        if not tags:
            return q

        user_settings = list(
            UserTagSetting.objects.filter(user=request.user, tag__in=tags).values_list(
                "name", flat=True
            )
        )
        if user_settings:
            q = q.filter(name__in=user_settings)

        combine = Q()
        for tag in tags:
            if tag in USER_TAGS:
                continue
            combine &= Q(tags__iregex=rf"(^|\n){tag}($|\n)")

        return q.filter(combine)


class ContentSettinForm(ModelForm):
    def __init__(self, *args, **kwargs):
        super(ContentSettinForm, self).__init__(*args, **kwargs)
        if self.instance.name:
            cs_type = get_type_by_name(self.instance.name)
            self.fields["value"] = cs_type.field
        self.fields["value"].strip = False

        if "user_defined_type" in self.fields:
            self.fields["user_defined_type"].widget = forms.Select(
                choices=[(v[0], v[2]) for v in USER_DEFINED_TYPES]
            )

    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)

    class Meta:
        model = ContentSetting
        fields = ["value"]


class ContentSettingAdmin(admin.ModelAdmin):
    list_display = ["name", "value", "setting_help", "setting_tags", "marks"]
    list_editable = [
        "value",
    ]
    search_fields = ["name", "value", "tags", "help"]
    readonly_fields = [
        "py_data",
        "name",
        "setting_help",
        "setting_tags",
        "version",
        "default_value",
        "fetch_one_setting",
    ]
    history_list_display = ["value"]
    form = ContentSettinForm

    def save_model(self, request, obj, form, change):
        if user_able_to_update(request.user, obj.name, obj.user_defined_type):
            super().save_model(request, obj, form, change)
            HistoryContentSetting.update_last_record_for_name(obj.name, request.user)
        else:
            add_message(
                request,
                ERROR,
                _("You are not allowed to change %(name)s") % {"name": obj.name},
            )

    def get_readonly_fields(self, request, obj=None):
        if not USER_DEFINED_TYPES or obj and not obj.user_defined_type:
            return super().get_readonly_fields(request, obj)

        return [
            v
            for v in super().get_readonly_fields(request, obj)
            if v
            not in (
                "name",
                "setting_help",
                "setting_tags",
            )
        ]

    def get_form(self, request, obj=None, change=False, **kwargs):
        if USER_DEFINED_TYPES and not obj or obj.user_defined_type:
            kwargs["fields"] = ["user_defined_type", "name", "value", "help", "tags"]
        else:
            kwargs["fields"] = [
                "value",
            ]

        return super().get_form(request, obj=None, change=False, **kwargs)

    def get_changelist(self, request, **kwargs):
        return SettingsChangeList

    def context_tags(self, request):
        extra_context = {}
        selected_tags = get_selected_tags_from_params(request.GET)

        def q_tags(tags):
            if not tags:
                return "?"
            return "?" + TAGS_PARAM + "=" + urllib.parse.quote(TAGS_SPLITER.join(tags))

        def q_add_tag(tag):
            return q_tags(selected_tags | set([tag]))

        def q_remove_tag(tag):
            return q_tags(selected_tags - set([tag]))

        extra_context["selected_tags"] = [
            {
                "tag": USER_TAGS[tag][0] if tag in USER_TAGS else tag,
                "url": q_remove_tag(tag),
            }
            for tag in sorted(selected_tags)
        ]

        tags_stat = defaultdict(int)
        user_settings = UserTagSetting.get_user_settings(request.user)
        for cs in ContentSetting.objects.all():
            val_tags = cs.tags_set | user_settings[cs.name]
            if selected_tags and (not val_tags or selected_tags - val_tags):
                continue

            for tag in val_tags - selected_tags:
                tags_stat[tag] += 1

        extra_context["available_tags"] = [
            {
                "tag": USER_TAGS[tag][0] if tag in USER_TAGS else tag,
                "url": q_add_tag(tag),
                "total": total,
            }
            for tag, total in sorted(
                tags_stat.items(), key=lambda a: a[0] and a[0] not in USER_TAGS
            )
        ]

        return extra_context

    def changelist_view(self, request, extra_context=None):
        return super().changelist_view(
            request, {**(extra_context or {}), **self.context_tags(request)}
        )

    def context_tags_view(self, request):
        return TemplateResponse(
            request,
            "admin/content_settings/contentsetting/context_tags.html",
            self.context_tags(request),
        )

    def history_view(self, request, object_id, extra_context=None):

        # First check if the user can see this history.
        model = self.model
        obj = self.get_object(request, unquote(object_id))
        if obj is None:
            return self._get_obj_does_not_exist_redirect(
                request, model._meta, object_id
            )

        if not self.has_view_or_change_permission(request, obj):
            raise PermissionDenied

        opts = model._meta
        app_label = opts.app_label

        context = {
            **self.admin_site.each_context(request),
            "title": f"Change history: {obj}",
            "subtitle": None,
            "module_name": str(capfirst(opts.verbose_name_plural)),
            "object": obj,
            "opts": opts,
            "preserved_filters": self.get_preserved_filters(request),
            "action_list": HistoryContentSetting.gen_unique_records(obj.name),
            **(extra_context or {}),
        }

        request.current_app = self.admin_site.name

        return TemplateResponse(
            request, f"admin/{app_label}/{opts.model_name}/object_history.html", context
        )

    def get_changelist_form(self, request, **kwargs):
        return ContentSettinForm

    def has_delete_permission(self, request, obj=None):
        default_del_permission = super().has_delete_permission(request, obj)
        if not default_del_permission or not USER_DEFINED_TYPES:
            return False
        return not obj or obj.user_defined_type is not None

    def has_add_permission(self, *args):
        return super().has_add_permission(*args) and USER_DEFINED_TYPES

    def has_change_permission(self, request, obj=None):
        default_change_permission = super().has_change_permission(request, obj)
        if not default_change_permission:
            return False
        if not obj:
            return True
        return user_able_to_update(request.user, obj.name)

    def setting_help(self, obj):
        cs_type = get_type_by_name(obj.name)
        if cs_type is None or cs_type.constant:
            return "(the setting is not using)"
        return mark_safe(obj.help)

    setting_help.short_description = "Help"

    def setting_tags(self, obj):
        cs_type = get_type_by_name(obj.name)
        if cs_type is None or cs_type.constant:
            return "(the setting is not using)"

        tags = obj.tags_set

        if not tags:
            return "---"
        return ", ".join(tags)

    setting_tags.short_description = "Tags"

    def marks(self, obj):
        return mark_safe(
            "".join(
                f"""<span>
            <span class="user_tag user_tag_selected user_tag_hidden" data-tag="{tag}" data-name="{obj.name}">{marks[0]}</span>
            <span class="user_tag user_tag_hidden" data-tag="{tag}" data-name="{obj.name}">{marks[1]}</span>
            </span>"""
                for tag, marks in USER_TAGS.items()
            )
        )

    def py_data(self, obj):
        cs_type = get_type_by_name(obj.name)

        if cs_type is None or cs_type.constant:
            return ""

        return mark_safe(cs_type.get_admin_preview_value(obj.value, obj.name))

    def default_value(self, obj):
        from django.utils.html import escape

        cs_type = get_type_by_name(obj.name)

        if cs_type is None or cs_type.constant:
            return ""

        attr_default = (
            cs_type.default.replace("&", "&amp;")
            .replace("\n", "&NewLine;")
            .replace('"', "&quot;")
        )
        return mark_safe(
            f"""
        <pre class="default_value">{escape(cs_type.default)}</pre>
        <a class="reset_default" data-value="{attr_default}" style="cursor: pointer; color:var(--link-fg)">Reset ("save" is required for applying)</a>
        """
        )

    def fetch_one_setting(self, obj):
        if not obj.name:
            return ""

        def get_fetch_reverse(name):
            return reverse(
                "content_settings:fetch_one_setting",
                kwargs={"name": name.lower().replace("_", "-")},
            )

        urls = [get_fetch_reverse(obj.name)]
        return "\n".join(urls)

    def get_urls(self):
        urls = super().get_urls()
        custom_urls = [
            path(
                "preview/",
                self.admin_site.admin_view(self.preview_setting),
                name="preview_setting",
            ),
            path("add-tag/", self.admin_site.admin_view(self.add_tag), name="add_tag"),
            path(
                "remove-tag/",
                self.admin_site.admin_view(self.remove_tag),
                name="remove_tag",
            ),
            path(
                "context-tags/",
                self.admin_site.admin_view(self.context_tags_view),
                name="context_tags",
            ),
        ]
        return custom_urls + urls

    # ajax view that creates UserTagSetting
    def add_tag(self, request):
        try:
            name = request.POST["name"]
            tag = request.POST["tag"]
            UserTagSetting.objects.create(
                user=request.user,
                name=name,
                tag=tag,
            )
        except Exception as e:
            return JsonResponse(
                {
                    "error": str(e),
                }
            )
        else:
            return JsonResponse(
                {
                    "success": True,
                }
            )

    # ajax view that removes UserTagSetting
    def remove_tag(self, request):
        try:
            name = request.POST["name"]
            tag = request.POST["tag"]
            UserTagSetting.objects.get(
                user=request.user,
                name=name,
                tag=tag,
            ).delete()
        except Exception as e:
            return JsonResponse(
                {
                    "error": str(e),
                }
            )
        else:
            return JsonResponse(
                {
                    "success": True,
                }
            )

    def preview_setting(self, request):
        if request.POST.get("user_defined_type"):
            try:
                cs_type_instance = USER_DEFINED_TYPES_INSTANCE[
                    request.POST["user_defined_type"]
                ]
            except KeyError:
                return JsonResponse(
                    {
                        "error": "Invalid user_defined_type",
                        "html": "",
                    }
                )
            else:
                cs_type = cs_type_instance()
        else:
            cs_type = get_type_by_name(request.POST["name"])
        if cs_type is None or cs_type.constant:
            return JsonResponse({"html": "", "error": "Invalid name"})

        value = request.POST["value"]

        other_values = {
            name[2:]: value
            for name, value in request.POST.items()
            if name.startswith("o_")
        }

        params = {
            name[2:]: value
            for name, value in request.POST.items()
            if name.startswith("p_")
        }

        try:
            cs_type.validate_value(value)
        except Exception as e:
            return JsonResponse(
                {
                    "error": str(e),
                }
            )
        with content_settings_context(**other_values):
            return JsonResponse(
                {
                    "html": cs_type.get_admin_preview_value(
                        value, request.POST["name"], **params
                    ),
                }
            )


class HistoryContentSettingAdmin(admin.ModelAdmin):
    list_display = [
        "name",
        "value",
        "user_defined_type",
        "version",
        "tags",
        "help",
        "was_changed",
        "user",
    ]
    list_filter = [
        "was_changed",
        "name",
    ]
    search_fields = ["name", "value", "tags", "help"]

    def has_delete_permission(self, request, obj=None):
        return False

    def has_add_permission(self, *args):
        return False

    def has_change_permission(self, request, obj=None):
        return False


admin.site.register(ContentSetting, ContentSettingAdmin)
admin.site.register(HistoryContentSetting, HistoryContentSettingAdmin)



from django import template
from content_settings.conf import content_settings
from django.utils.html import escape

register = template.Library()


@register.simple_tag
def content_settings_call(name, *args, **kwargs):
    safe = kwargs.pop("_safe", False)
    result = getattr(content_settings, name)(*args, **kwargs)
    return result if safe else escape(result)



# some job urls declared in api/urls.py
from django.urls import path
from .views import fetch_one_setting

app_name = "content_settings"
urlpatterns = [
    path("fetch/<str:name>/", fetch_one_setting, name="fetch_one_setting"),
    path(
        "fetch/<str:name>/suffix/<str:suffix>/",
        fetch_one_setting,
        name="fetch_suffix_one_setting",
    ),
]



from django import template
from content_settings.conf import content_settings
from django.utils.html import escape

from content_settings.models import UserTagSetting

register = template.Library()


@register.simple_tag(takes_context=True)
def content_settings_user_tags(context, *args, **kwargs):
    user = context["request"].user
    return list(UserTagSetting.objects.filter(user=user).values_list("name", "tag"))



from inspect import isclass
from functools import cached_property
from pprint import pformat
from typing import Optional, Set, Tuple, Union, Any, Callable
from collections.abc import Iterable

from django import forms

from content_settings.context_managers import context_defaults_kwargs
from content_settings.settings import CACHE_SPLITER
from content_settings.types.lazy import LazyObject
from content_settings.types import (
    PREVIEW_ALL,
    PREVIEW_HTML,
    PREVIEW_TEXT,
    PREVIEW_PYTHON,
    PREVIEW_NONE,
)


class BaseSetting:
    pass


class SimpleString(BaseSetting):
    """
     Attributes:
    - constant (bool): Whether the setting is constant (can not be changed).
    - cls_field (forms.CharField): The form field class to use for the setting.
    - widget (forms.Widget): The form widget to use for the cls_field.
    - widget_attrs (Optional[dict]): Optional attributes for the widget initiation.
    - fetch_permission (Optional[str]): Optional permission required to fetch the setting.
    - update_permission (Optional[str]): Optional permission required to update the setting.
    - help_format (Optional[str]): Optional format string for the help text for the format (align to the type).
    - help (Optional[str]): Optional help text for the setting.
    - value_required (bool): Whether a value is required for the setting.
    - version (str): The version of the setting (using for caching).
    - tags (Optional[Iterable[str]]): Optional tags associated with the setting.
    - validators (Tuple[Callable]): Validators to apply to the setting value.
    - empty_is_none (bool): Whether an empty value should be treated as None.
    - admin_preview_as (str): The format to use for the admin preview.
    - suffixes (Tuple[str]): Suffixes that can be appended to the setting value.
    - user_defined_slug (str): it contains a slug from db If the setting is defined in DB only (should not be set in content_settings)
    - overwrite_user_defined (bool): Whether the setting can overwrite a user defined setting.
    - default (str): The default value for the setting.
    """

    constant: bool = False
    cls_field: forms.CharField = forms.CharField
    widget: forms.Widget = forms.TextInput
    widget_attrs: Optional[dict] = None
    fetch_permission: Optional[str] = None
    update_permission: Optional[str] = None
    help_format: Optional[str] = ""
    help: Optional[str] = None
    value_required: bool = False
    version: str = ""
    tags: Optional[Iterable[str]] = None
    validators: Tuple[Callable] = ()
    empty_is_none: bool = False
    admin_preview_as: str = PREVIEW_NONE
    suffixes: Tuple[str] = ()
    user_defined_slug: Optional[str] = None
    overwrite_user_defined: bool = False
    default: str = ""

    def __init__(self, default: Optional[str] = None, **kwargs):
        for k in kwargs.keys():
            assert self.can_assign(k), "Attribute {} not found".format(k)

        if default is not None:
            self.default = default
        assert isinstance(self.default, str), "Default should be str"

        kwargs = self.update_defaults_context(kwargs)
        self.init_assign_kwargs(kwargs)

        assert isinstance(self.version, str), "Version should be str"
        assert (
            CACHE_SPLITER not in self.version
        ), f"Version should not contain CACHE_SPLITER:{CACHE_SPLITER}"
        assert (
            self.get_admin_preview_as() in PREVIEW_ALL
        ), f"admin_preview_as should be in {PREVIEW_ALL}"

    def get_admin_preview_as(self) -> str:
        return self.admin_preview_as

    def init_assign_kwargs(self, kwargs):
        for k, v in kwargs.items():
            if not self.can_assign(k):
                raise ValueError("Attribute {} not found".format(k))
            if hasattr(self, "init__{}".format(k)):
                getattr(self, "init__{}".format(k))(v)
            else:
                setattr(self, k, v)

    def init__tags(self, tags: Union[None, str, Iterable[str]]) -> None:
        if not tags:
            return

        if isinstance(tags, str):
            tags = {tags}
        else:
            tags = set(tags)

        self.tags = tags if self.tags is None else self.tags | tags

    @cached_property
    def field(self) -> forms.Field:
        return self.get_field()

    def get_suffixes(self) -> Tuple:
        return self.suffixes

    def can_suffix(self, suffix: Optional[str]) -> bool:
        return suffix is None or suffix in self.get_suffixes()

    def can_assign(self, name: str) -> bool:
        return hasattr(self, name)

    def update_defaults_context(self, kwargs: dict) -> dict:
        # initiate default values
        context_defaults = {
            name: value
            for name, value in context_defaults_kwargs().items()
            if self.can_assign(name)
        }

        # update actual values
        return {
            name: value
            for name, value in context_defaults_kwargs(
                {
                    **context_defaults,
                    **kwargs,
                }
            ).items()
            if self.can_assign(name)
        }

    def get_help_format(self) -> Optional[Union[str, Iterable[str]]]:
        yield self.help_format

    def get_help(self) -> Optional[str]:
        help_format = "".join(self.get_help_format())
        if help_format:
            return f"{self.help} <br><br> {help_format}"
        return self.help

    def get_tags(self) -> Set[str]:
        tags = self.tags
        if tags is None:
            tags = set()
        return set(tags)

    def get_validators(self) -> Tuple[Callable]:
        return self.validators + tuple(self.cls_field.default_validators)

    def get_field(self) -> forms.Field:
        return self.cls_field(
            widget=self.get_widget(),
            validators=(self.validate_value,),
            required=self.value_required,
        )

    def get_widget(self) -> forms.Widget:
        if isclass(self.widget) and self.widget_attrs is not None:
            return self.widget(attrs=self.widget_attrs)
        else:
            return self.widget

    def validate_raw_value(self, value: str) -> None:
        pass

    def validate_value(self, value: str) -> Any:
        self.validate_raw_value(value)
        val = self.to_python(value)
        self.validate(val)

    def validate(self, value):
        for validator in self.get_validators():
            validator(value)

    def to_python(self, value: str) -> Any:
        if self.empty_is_none and value.strip() == "":
            return None
        return self.field.to_python(value)

    def json_view_value(self, request, value: Any) -> Any:
        return value

    def give_python_to_admin(self, value: str, name: str, **kwargs) -> Any:
        return self.to_python(value)

    def get_admin_preview_html(self, value: str, name: str, **kwargs) -> Any:
        return self.give_python_to_admin(value, name, **kwargs)

    def get_admin_preview_text(self, value: str, name: str, **kwargs) -> Any:
        return self.give_python_to_admin(value, name, **kwargs)

    def get_admin_preview_python(self, value: str, name: str, **kwargs) -> Any:
        return self.give_python_to_admin(value, name, **kwargs)

    def get_admin_preview_value(self, value: str, name: str, **kwargs) -> str:
        if self.get_admin_preview_as() == PREVIEW_NONE:
            return ""

        if self.get_admin_preview_as() == PREVIEW_HTML:
            return str(self.get_admin_preview_html(value, name, **kwargs))

        if self.get_admin_preview_as() == PREVIEW_TEXT:
            value = str(self.get_admin_preview_text(value, name, **kwargs))
        else:
            value = pformat(self.get_admin_preview_python(value, name, **kwargs))

        return "<pre>{}</pre>".format(
            value.replace("<", "&lt;")
            .replace(">", "&gt;")
            .replace("&", "&amp;")
            .replace('"', "&quot;")
        )

    def lazy_give(self, l_func: Callable, suffix=None) -> LazyObject:
        return LazyObject(l_func)

    def give(self, value, suffix: Optional[str] = None):
        return value

    def give_python(self, value: str, suffix=None) -> Any:
        return self.give(self.to_python(value), suffix)


class SimpleText(SimpleString):
    widget: forms.Widget = forms.Textarea
    widget_attrs: dict = {"rows": 10, "cols": 80}


class SimpleHTML(SimpleText):
    admin_preview_as: str = PREVIEW_HTML


class URLString(SimpleString):
    cls_field: forms.Field = forms.URLField
    widget: forms.Widget = forms.URLInput
    widget_attrs: dict = {"style": "max-width: 600px; width: 100%"}


class EmailString(SimpleString):
    cls_field: forms.Field = forms.EmailField
    widget: forms.Widget = forms.EmailInput
    widget_attrs: dict = {"style": "max-width: 600px; width: 100%"}


class SimpleInt(SimpleString):
    admin_preview_as: str = PREVIEW_PYTHON
    cls_field: forms.Field = forms.IntegerField

    def get_help_format(self):
        yield "Any number"


class SimpleBool(SimpleInt):
    admin_preview_as: str = PREVIEW_PYTHON
    min_value: int = 0
    max_value: int = 1
    empty_is_none: bool = True

    def to_python(self, value) -> bool:
        return bool(super().to_python(value))


class SimpleDecimal(SimpleString):
    admin_preview_as: str = PREVIEW_PYTHON
    cls_field: forms.Field = forms.DecimalField



from itertools import zip_longest

from django.core.exceptions import ValidationError

from .basic import SimpleText
from .mixins import CallToPythonMixin, GiveCallMixin
from .validators import call_validator
from . import PREVIEW_TEXT
from ..permissions import superuser


class required:
    pass


class SimpleCallTemplate(CallToPythonMixin, SimpleText):
    admin_preview_as = PREVIEW_TEXT
    template_static_includes = ("CONTENT_SETTINGS", "SETTINGS")
    template_static_data = None
    template_args_default = None
    help_format = "Simple <a href='https://docs.djangoproject.com/en/3.2/topics/templates/' target='_blank'>Django Template</a>. Available objects:"

    def prepare_python_call(self, value):
        raise NotImplementedError()

    def python_call(self, *args, **kwargs):
        raise NotImplementedError()

    def get_template_args_default(self):
        if not self.template_args_default:
            return {}
        return self.template_args_default

    def get_template_static_data(self):
        if not self.template_static_data:
            return {}
        if callable(self.template_static_data):
            return self.template_static_data()
        return self.template_static_data

    def get_help_format(self):
        yield self.help_format
        yield "<ul>"

        for name in self.get_template_args_default().keys():
            yield f"<li>{name}</li>"

        for name in self.get_template_static_data().keys():
            yield f"<li>{name}</li>"

        yield "</ul>"

    def prepate_input_to_dict(self, *args, **kwargs):
        template_args_default = self.get_template_args_default()
        no_value = {}
        kwargs = {
            **{
                default[0]: (default[1] if arg_value == no_value else arg_value)
                for arg_value, default in zip_longest(
                    args, template_args_default.items(), fillvalue=no_value
                )
                if default != no_value
            },
            **kwargs,
        }

        for name, value in kwargs.items():
            if value == required:
                raise ValidationError(f"Missing required argument {name}")

        return kwargs

    def get_template_full_static_data(self):
        ret = {}

        if "CONTENT_SETTINGS" in self.template_static_includes:
            from content_settings.conf import content_settings

            ret["CONTENT_SETTINGS"] = content_settings

        if "SETTINGS" in self.template_static_includes:
            from django.conf import settings

            ret["SETTINGS"] = settings

        return {
            **ret,
            **self.get_template_static_data(),
        }


class DjangoTemplate(SimpleCallTemplate):
    tags = {"template"}

    def prepare_python_call(self, value):
        from django.template import Template

        return {"template": Template(value.strip())}

    def python_call(self, *args, **kwargs):
        from django.template import Context

        template = kwargs.pop("template")
        return template.render(
            Context(
                {
                    **self.get_template_full_static_data(),
                    **self.prepate_input_to_dict(*args, **kwargs),
                }
            )
        )


class DjangoTemplateNoArgs(GiveCallMixin, DjangoTemplate):
    pass


class DjangoModelTemplate(DjangoTemplate):
    model_queryset = None
    obj_name = "object"
    admin_preview_call = False

    def get_template_args_default(self):
        return {
            self.obj_name: required,
            **super().get_template_args_default(),
        }

    def get_first_call_validator(self):
        if self.model_queryset is not None:
            first = self.model_queryset.first()
            if first is not None:
                return call_validator(first)

    def get_validators(self):
        validators = super().get_validators()
        first_validator = self.get_first_call_validator()

        if first_validator is not None:
            validators = (first_validator, *validators)

        return validators

    def get_preview_validators(self):
        first_validator = self.get_first_call_validator()

        if first_validator is not None:
            return (first_validator,)

        return ()


class SimpleEval(SimpleCallTemplate):
    update_permission = superuser
    help_format = "Python code that returns a value"
    tags = {"eval"}

    def prepare_python_call(self, value):
        return {"template": compile(value, "<string>", "eval")}

    def python_call(self, *args, **kwargs):
        template = kwargs.pop("template")
        globs = {
            **self.get_template_full_static_data(),
            **self.prepate_input_to_dict(*args, **kwargs),
            "__import__": None,
        }

        return eval(template, globs)


class SimpleEvalNoArgs(GiveCallMixin, SimpleEval):
    pass


class SimpleExec(SimpleCallTemplate):
    update_permission = superuser
    help_format = "Python code that execute and returns generated variables "
    tags = {"eval"}
    call_return = None

    def get_call_return(self):
        if self.call_return is None:
            return None
        if isinstance(self.call_return, dict):
            return self.call_return

        return {name: None for name in self.call_return}

    def get_help_format(self):
        yield from super().get_help_format()
        yield f"Expected variables: <ul>"
        for name, value in self.get_call_return().items():
            yield f"<li>{name}: {value}</li>"
        yield "</ul>"

    def prepare_python_call(self, value):
        return {"template": compile(value, "<string>", "exec")}

    def python_call(self, *args, **kwargs):
        template = kwargs.pop("template")
        globs = {
            **self.get_template_full_static_data(),
            **self.prepate_input_to_dict(*args, **kwargs),
            "__import__": None,
        }

        exec(template, globs)
        call_return = self.get_call_return()

        if call_return is None:
            return globs
        return {k: globs.get(k, v) for k, v in call_return.items()}


class SimpleExecNoArgs(GiveCallMixin, SimpleExec):
    pass



from django.core.exceptions import ValidationError
from pprint import pformat


class call_validator:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __call__(self, func):
        try:
            return func(*self.args, **self.kwargs)
        except Exception as e:
            raise ValidationError(str(e))

    def __str__(self) -> str:
        ret = ""
        if self.args:
            ret += ", ".join([pformat(arg) for arg in self.args])

        if self.kwargs:
            if ret:
                ret += ", "
            ret += ", ".join(
                [f"{key}={pformat(value)}" for key, value in self.kwargs.items()]
            )

        return ret

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {str(self)}>"
